import subprocess
import sys
import os
from elftools.elf.elffile import ELFFile
from elftools.elf.sections import SymbolTableSection 
import dbus  
import configparser

PATH_LIST=["/bin/", "/usr/bin/", "/sbin/", "/usr/sbin/"]
shells = {
    #key:value
    #cmd:[]
}

def install_shell_package():
    for cmd_list in shells.values():
        for cmd in cmd_list:
            packages = subprocess.getoutput('apt-file search {}'.format(cmd))
            ff = False
            if packages:
                packages = packages.split("\n")
                for package in packages:
                    if cmd == package.split(": ")[1]:
                        os.system("apt install {} -y".format(package.split(": ")[0]))
                        ff = True
                        break
            if ff:
                break

# 指定要检测的 Shell 脚本路径
def check_shell(path, debian=False):
    error_message = None
    try:  
        process = subprocess.run([path], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)  
        if process.stderr.decode():
            error_message= process.stderr.decode()
            # install_shell_package(error_message=)
    except subprocess.CalledProcessError as e:
            error_message= e.stderr.decode()
    if error_message:
        errors = error_message.split("\n")
        for err in errors:
            cmd = None
            if "not found" in err :
                cmd = err.split(": ")[-2]
            elif  "未找到命令"  in err:
                cmd = err.split("：")[-2].split(": ")[-1]
            if debian and cmd:
                return cmd
            if cmd and cmd not in shells.keys():
                [ shells.setdefault(cmd, []).append(i+cmd) for i in PATH_LIST]


def find_post_shell(file_path, cmd):
    if os.path.isdir(file_path):
        for path in os.listdir(file_path):
            find_post_shell(file_path+"/"+path, cmd)
    if os.path.isfile(file_path):
        if cmd == file_path.split("/")[-1]:
            return True
    return False
    

def find_shell(file_path, debian=False):
    for path in os.listdir(file_path):
        if "postinst" in path:
            cmd = check_shell(file_path+"/"+path, debian)
            if find_post_shell(file_path.rsplit("/", 2)[0]+"/usr", cmd) is False:
                [ shells.setdefault(cmd, []).append(i+cmd) for i in PATH_LIST]


def dpkg_get_exec_depends(need):
    try:
        outputs = subprocess.check_output(['dpkg', '-S', need], stderr=subprocess.STDOUT).decode('utf-8').split("\n")
    except subprocess.CalledProcessError:
        return None
    for deb in outputs:
        if "/usr/lib" in deb or "/lib" in deb:
            need_path = deb.split(":")[1].strip()
            if need_path.startswith("/") is False:
                need_path = deb.split(":")[2].strip()
            if need_path.startswith("/usr/lib") or need_path.startswith("/lib"):
                return deb.split(":")[0]

def apt_file_get_depends(need):
    depends = subprocess.getoutput('apt-file search {}'.format(need)).split("\n")
    for depend in depends:
        arch = subprocess.getoutput("arch")
        if arch not in depend:
            continue
        if "/usr/lib" in depend or "/lib" in depend:
            need_path = depend.split(":")[1].strip()
            if need_path.startswith("/") is False:
                need_path = depend.split(":")[2].strip()
            if need_path.startswith("/usr/lib") or need_path.startswith("/lib"):
                return depend.split(":")[0]

# 检测二进制
def check_exec(path):
    # 检测文件是否为ELF二进制文件
    output = subprocess.check_output(['file', path]).decode('utf-8')
    if output.split(path+":")[-1].strip().startswith("ELF") is False:
        try:
            with open(path, "r") as f:
                first_line = f.readline().strip()
                if first_line == "#!/bin/sh" or first_line == "#!/bin/bash" or first_line == "#!/usr/bin/bash"  or first_line == "#!/usr/bin/sh":
                    f.close()
                    check_shell(path)
        except UnicodeDecodeError as e:
            f.close()
        return
    # elf_name = path.split("/")[-1]
    # if "so" in elf_name.split("."):
    #     return

    # 检测所需依赖
    output = subprocess.check_output(['objdump', '-p', path]).decode('utf-8')
    dependencies, needed = [], []

    # 解析 objdump 输出，提取依赖项
    lines = output.split('\n')
    for line in lines:
        if "NEEDED" in line:
            dependency = line.split("NEEDED")[1].strip()
            if dpkg_get_exec_depends(dependency) is None:
                dependencies.append(dependency)
    if len(dependencies) == 0:
        return
    for need in dependencies:
        lib = apt_file_get_depends(need)
        if lib is None:
            continue
        needed.append(lib)
    # TODO dlopen部分未解决
    return needed

def install_depend(depends):
    # print(depends)
    os.system("apt install {} -y".format(" ".join(depends)))

def find_exec(file_path):
    if os.path.isdir(file_path):
        for path in os.listdir(file_path):
            find_exec(file_path+"/"+path)
    if os.path.isfile(file_path):
        depends = check_exec(file_path)
        try:
            if len(depends) != 0 :
                install_depend(depends)
        except TypeError as e:
            pass


def unpack_deb(file, debian=False):
    unpack_path = "unpack-{}".format(file.rsplit("/")[-1])
    if debian:
        subprocess.check_output(['dpkg', '-X', file, unpack_path]).decode('utf-8')
        subprocess.check_output(['dpkg', '-e', file, unpack_path+'/DEBIAN']).decode('utf-8')
        find_shell(unpack_path + "/DEBIAN", True)
        install_shell_package()
        sys.exit()

    find_exec(unpack_path + "/usr")
    find_exec(unpack_path + "/opt")
    install_shell_package()
    os.system("rm -rf {}".format(unpack_path))

def call_dbus(command):
    bus = dbus.SystemBus()  
    service_name = 'com.kylin.kare.host'  
    object_path = '/com/kylin/kare/host'  
    interface_name = 'com.kylin.kare.host'  
    proxy_obj = bus.get_object(service_name, object_path)
    interface = dbus.Interface(proxy_obj, interface_name)
    result = interface.command(command)
    if result:
        if os.path.exists(result):
            return
        f = open(result, "w")
        # f.write()
        f.write(r"""
#!/bin/bash
params="{}" 
              
if [ $# -gt 0 ]; then  
    for arg in "$@"; do  
        if [ -n "$params" ]; then  
            params="$params "  
        fi
        params="$params$arg"  
    done  
    echo $params
fi

dbus-send --session --dest=com.kylin.kare.Sessionhost --type=method_call --print-reply /com/kylin/kare/Sessionhost com.kylin.kare.Sessionhost.call string:"$params"
                """.format(result.rsplit("/", 1)[-1]))
        f.close()
        subprocess.getoutput("chmod +x {}".format(result))
        print(result)

def check_for_system_call(elf_path):
    with open(elf_path, 'rb') as f:
        elf = ELFFile(f)
        for section in elf.iter_sections():
            if isinstance(section, SymbolTableSection):
                for symbol in section.iter_symbols():
                    if symbol.name == 'system' or symbol.name == b'system':
                        return True 
        dynsym_section = elf.get_section_by_name('.dynsym')
        if dynsym_section:
            for symbol in dynsym_section.iter_symbols():
                if symbol.name == 'system':
                    return True
    return False

def contains_non_ascii(s):
    return all(96<ord(c)<123 or 64<ord(c)<91 or 44<ord(c)<58  for c in s)

def read_elf_file(file_path):
    # if check_for_system_call(file_path) is False:
    #     return
    with open(file_path, 'rb') as f:
        elf_file = ELFFile(f)
        for section in elf_file.iter_sections():
            if section.name == '.rodata' or section.name == b'.rodata':
                data = section.data()
                string_data = data.decode('utf-8', 'ignore').split('\0')
                for s in string_data:
                    if contains_non_ascii(s) and len(s) > 1:
                        if len(subprocess.getoutput("command -v {}".format(s))) == 0:
                            call_dbus(s)


def read_python_file(file_path):

    if subprocess.getoutput("strings {} | grep os.system"):
        f = open(file=file_path, mode="r")
        for i in f.readlines():
            if "os.system" in i and not i.strip().startswith("#"):
                for x in i.strip().split("os.system")[-1].split("(")[1:]:
                    x = x.replace("\"", "")
                    x = x.replace("\'", "")
                    for cmd in x.split(" "):
                        if contains_non_ascii(cmd) and len(cmd) > 1:
                            call_dbus(cmd)

    if subprocess.getoutput("strings {} | grep os.subprocess"):
        f = open(file=file_path, mode="r")
        for i in f.readlines():
            if "subprocess." in i and not i.strip().startswith("#"):
                for x in i.strip().split("subprocess.")[-1].split("(")[1:]:
                    x = x.replace("\"", "")
                    x = x.replace("\'", "")
                    for cmd in x.split(" "):
                        if contains_non_ascii(cmd) and len(cmd) > 1:
                            call_dbus(cmd)


def find_desktop_exec(file):
    desktop_maps = {}
    unpack_path = "unpack-{}".format(file.rsplit("/")[-1])
    subprocess.check_output(['dpkg', '-X', file, unpack_path]).decode('utf-8')
    desktops = subprocess.getoutput("dpkg -c {} | grep desktop$".format(file)).split("\n")
    for desktop in desktops:
        desktop_maps.setdefault(desktop.rsplit("/")[-1], unpack_path+desktop.split(" ")[-1].strip(".")) 
    
    for desktop_path in desktop_maps.values():
        config = configparser.ConfigParser()
        config.read(desktop_path)  
        default_section = config['Desktop Entry']
        exec_value = default_section.get('Exec')  
        output = subprocess.check_output(['file', unpack_path+exec_value]).decode('utf-8')
        if output.split(unpack_path+exec_value+":")[-1].strip().startswith("ELF") is True:
            read_elf_file(unpack_path+exec_value)
        else:
            read_python_file(unpack_path+exec_value)

try:
    if sys.argv[2] == "1": 
        unpack_deb(sys.argv[1], True)
    elif sys.argv[2] == "2": 
        find_desktop_exec(sys.argv[1])
except:
    unpack_deb(sys.argv[1])