import os
import shutil
import logging
import threading
import subprocess
from gettext import gettext as _
from datetime import datetime, timedelta
from time import sleep
from RevisionsManager.Core.RMConf import RMConf
from RevisionsManager.Core.RMCom import RMCom, get_sys_call

TMP_FS              = '/.fs.old-rm/'
CHECK_PY_PATH       = '/usr/share/revisions-manager/RevisionsManager/Checkpkg/src/main.py'
STEP3_USED_TIME     = 10
STEP1_RECORD_TIME   = 24
STEP2_RECORD_TIME   = 48
STEP3_RECORD_TIME   = 72
STEP1_LIST_FILE     = "/usr/share/revisions-manager/data/.oldfs-removelist"  # 必定无用名单
STEP2_LIST_FILE     = "/kpct.list"   # 白名单

def update_grub():
    logging.info("[RMCOM]: update grub.")
    if os.path.exists("/usr/sbin/update-grub2"):
        os.system("update-grub2")
    elif os.path.exists("/usr/sbin/update-grub"):
        os.system("update-grub")
    else:
        logging.error("Base tools not exists [update-grub]")
        return -1
    return 0

class RMClear():
    STAGE_B = 0
    STAGE_1_B = 1
    STAGE_1_E = 2
    STAGE_2_B = 3
    STAGE_2_E = 4
    STAGE_3_B = 5
    STAGE_E = 6

    SEC_CLEAR = 'clear'
    KEY_RECORDINTREVAL = 'recordinterval'
    KEY_AUTOCLEARINTERVAL = 'clearinterval'
    KEY_STAGESW = 'stageswitch'
    KEY_STAGEINTREVAL = 'stageinterval'

    def __init__(self):
        self.rmCom      = RMCom()
        self.uptime     = None
        self.period     = 30
        self.sw         = False
        self.used       = 0              # 开机次数统计
        self.flag       = False
        self.timer      = None
        self.rTime      = 0              # 使用时间统计 600 秒记录一次
        self.stage      = self.STAGE_B
        self.osName     = None

        self.recordInterval     = 600           # 600 秒     # 2
        self.autoClearInterval  = 3600 * 24     # 24 小时    # 20
        self.stageSw            = False         # n          # y
        self.stageInterval      = 60                         # 秒

        self.rmState = RMConf(self.rmCom.RM_STATE_PATH)
        self.rmConf = RMConf(self.rmCom.CONF_PATH)

        if self.is_cleared():
            logging.info("[RMCLEAR] is cleared.")
            return
        if not self.is_up():
            return

        self.get_oldfsName()
        self.get_test_param()
        self.get_clear_stage()
        self.add_use_time()
        self.get_switch()
        self.get_uptime()
        self.get_period()
        threadA = threading.Thread(target=self.auto_clear)
        threadA.start()
        threadB = threading.Thread(target=self.record_time)
        threadB.start()
        threadC = threading.Thread(target=self.check_pkg_file)
        threadC.start()
    
    def get_oldfsName(self):
        osName = self.rmState.get_key(self.rmCom.SEC_OS_INFO, self.rmCom.KEY_OLD_FS_NAME)
        if osName == None or not os.path.exists(osName):
            logging.error("[RMCLEAR] old fs [%s] not exists.", osName)
            return False
        if not osName.startswith("/"):
            logging.error("[RMCLEAR] old fs [%s] not right.", osName)
            return False            
        self.osName = osName
        return True

    def check_pkg_file(self):
        if not os.path.exists(CHECK_PY_PATH):
            logging.error("[RMCLEAR] [%s] not exists.", CHECK_PY_PATH)
            return

        if self.osName == None:
            logging.error("[RMCLEAR] check pkg file oldfs not exists.")
            return

        pkgPath = self.osName + RMCom.CACHE_DIR + RMCom.COMPAT_INFO_FILE
        if not os.path.exists(pkgPath):
            logging.error("[RMCLEAR] [%s] not exists.", pkgPath)
            return

        infoList = []
        with open(pkgPath, 'r') as fd:
            infoList = fd.readlines()
        fd.close()

        if len(infoList) <= 0:
            logging.info("[RMCLEAR] there is no info.")
            return

        for everyline in infoList:
            eleList = everyline.split("_")
            pkgName = eleList[0]
            cmd = "python3 " + CHECK_PY_PATH + " -d " + pkgName + " -fsold " + self.osName
            get_sys_call(cmd)

        return

    def get_test_param(self):
        try:
            recordIntervalS = self.rmConf.get_key(self.SEC_CLEAR, self.KEY_RECORDINTREVAL)
            if recordIntervalS != None:
                self.recordInterval = int(recordIntervalS)
        except Exception as e:
            logging.error("[RMCLEAR]: get [%s] err [%s]", self.KEY_RECORDINTREVAL, e)
        try:
            autoClearIntervalS = self.rmConf.get_key(self.SEC_CLEAR, self.KEY_AUTOCLEARINTERVAL)
            if autoClearIntervalS != None:
                self.autoClearInterval = int(autoClearIntervalS)
        except Exception as e:
            logging.error("[RMCLEAR]: get [%s] err [%s]", self.KEY_AUTOCLEARINTERVAL, e)
        try:
            stageSwS = self.rmConf.get_key(self.SEC_CLEAR, self.KEY_STAGESW)
            if stageSwS != None and stageSwS == 'y':
                self.stageSw = True
        except Exception as e:
            logging.error("[RMCLEAR]: get [%s] err [%s]", self.KEY_STAGESW, e)

        try:
            stageIntervalS = self.rmConf.get_key(self.SEC_CLEAR, self.KEY_STAGEINTREVAL)
            if stageIntervalS != None:
                self.stageInterval = int(stageIntervalS)
        except Exception as e:
            logging.error("[RMCLEAR]: get [%s] err [%s]", self.KEY_STAGEINTREVAL, e)

    def set_clear_state(self, state):
        if state:
            stateStr = "True"
        else:
            stateStr = "False"
        try:
            self.rmState.change_key_value(self.rmCom.SEC_STATE, self.rmCom.KEY_CLEAR, stateStr)
            logging.info("[RMCLEAR]: clear state change to [%s].", stateStr)
        except Exception as e:
            logging.error("[RMCLEAR] set  clear state err [%s].", e)

    def is_cleared(self):
        clearState = self.rmState.get_key(self.rmCom.SEC_STATE, self.rmCom.KEY_CLEAR)
        if clearState != None and clearState == "True":
            logging.info("[RMCLEAR]: clearState is True")
            return True
        else:
            logging.info("[RMCLEAR]: clearState is False")
            return False

    def reset_record_time(self):
        try:
            self.rmState.change_key_value(self.rmCom.SEC_STATE, self.rmCom.KEY_RUNNINGTIME, "0")
            self.rTime = 0
        except Exception as e:
                logging.error("[RMCLEAR]: reset record time err[%s].", e)

    def record_time(self):
        if self.stage >= self.STAGE_E:
            logging.info("[RMCLEAR] clear is over, record time stop.")
            return

        rTimeNew = "0"
        rTimeS = self.rmState.get_key(self.rmCom.SEC_STATE, self.rmCom.KEY_RUNNINGTIME)
        if rTimeS != None:
            try:
                rTimeInt = int(rTimeS)
                self.rTime = rTimeInt + 1
                rTimeNew = str(self.rTime)
            except Exception as e:
                logging.error("[RMCLEAR]: record time err[%s].", e)

        self.rmState.change_key_value(self.rmCom.SEC_STATE, self.rmCom.KEY_RUNNINGTIME, rTimeNew)

        timer = threading.Timer(self.recordInterval, function=self.record_time)
        timer.start()

    def reset_use_time(self):
        try:
            self.rmState.change_key_value(self.rmCom.SEC_STATE, self.rmCom.KEY_USED, "0")
            self.used = 0
        except Exception as e:
                logging.error("[RMCLEAR]: reset used time err[%s].", e)

    def add_use_time(self):
        tusedNew = "0"
        tused = self.rmState.get_key(self.rmCom.SEC_STATE, self.rmCom.KEY_USED)
        if tused != None:
            try:
                tusedInt = int(tused)
                self.used = tusedInt + 1
                tusedNew = str(self.used)
            except Exception as e:
                logging.error("[RMCLEAR]: get used time err[%s].", e)

        self.rmState.change_key_value(self.rmCom.SEC_STATE, self.rmCom.KEY_USED, tusedNew)

    def is_up(self):
        vmState = self.rmState.get_key(self.rmCom.SEC_STATE, self.rmCom.KEY_VM)
        if vmState == self.rmCom.intToStr.get(self.rmCom.VMUP):
            return True

        if vmState == self.rmCom.intToStr.get(self.rmCom.VMUPSUCCESS):
            return True

        logging.info("[RMCLEAR] state[%s] is not upover/upsuccess.", vmState)
        return False

    def auto_clear(self):
        if not self.sw:
            logging.info("[RMCLEAR] switch is false.")
            return
        if self.stage >= self.STAGE_E:
            logging.info("[RMCLEAR] clear is over, auto clear stop.")
            return
        try:
            self.update_stage()
            self.clear_act()
        except Exception as e:
            logging.error("[RMCLEAR] auto clear err [%s].", e)

        self.timer = threading.Timer(self.autoClearInterval, function=self.auto_clear)
        self.timer.start()

    def reset_clear_stage(self):
        try:
            self.rmState.change_key_value(self.rmCom.SEC_STATE, self.rmCom.KEY_CLEARSTAGE, "0")
            self.stage = 0
        except Exception as e:
                logging.error("[RMCLEAR]: reset clear stage err[%s].", e)

    def get_clear_stage(self):
        try:
            clearstage = self.rmState.get_key(self.rmCom.SEC_STATE,self.rmCom.KEY_CLEARSTAGE)
            if clearstage == None or len(clearstage) == 0:
                logging.warning("[RMCLEAR]: there is no clearstage")
                return
            self.stage = int(clearstage)
        except Exception as e:
            logging.error("[RMCLEAR] get clearstage err [%s].", e)


    # 获取升级时间
    def get_uptime(self):
        if self.uptime is not None:
            logging.info("[RMCLEAR] uptime is known [%s].", self.uptime)
            return
        try:
            fmt = "%Y-%m-%d %H:%M:%S"
            uptime = self.rmState.get_key(self.rmCom.SEC_TIME,self.rmCom.KEY_UPTIME)
            if uptime is None:
                logging.error("[RMCLEAR]: there is no uptime")
                return
            self.uptime = datetime.strptime(uptime,fmt)
        except Exception as e:
            logging.error("[RMCLEAR] get uptime err [%s].", e)

    def get_switch(self):
        try:
            swClear = self.rmConf.get_key(self.rmConf.SEC_SWITCH, self.rmConf.KEY_SW_CLEAR)
            if swClear is None:
                logging.info("[RMCLEAR]: there is no swClear")
            elif swClear == "True":
                self.sw = True
                logging.info("[RMCLEAR]: swClear is True")
            else:
                logging.info("[RMCLEAR]: swClear is False")
        except Exception as e:
            logging.error("[RMCLEAR] get switch err [%s].", e)
    
    def set_switch(self, state):
        if state:
            stateStr = "True"
        else:
            stateStr = "False"

        try:
            self.rmConf.change_key_value(self.rmConf.SEC_SWITCH, self.rmConf.KEY_SW_CLEAR, stateStr)
            self.sw = state
            logging.info("[RMCLEAR]: swClear change to [%s].", stateStr)
        except Exception as e:
            logging.error("[RMCLEAR] set switch err [%s].", e)        

    # 获取间隔时间
    def get_period(self):
        if self.period is not None:
            logging.info("[RMCLEAR] period is known [%d].", self.period)
            return
        try:
            periodstr = self.rmConf.get_key(self.rmConf.SEC_PERIOD, self.rmConf.KEY_BACKPERIOD)
            if periodstr is None:
                logging.error("[RMCLEAR]: there is no period")
                return
            self.period = int(periodstr)
        except Exception as e:
            logging.error("[RMCLEAR] get period err [%s].", e)

    def update_stage(self):
        if self.uptime is None:
            logging.error("[RMCLEAR]: uptime not exists.")
            return
        if self.period is None:
            self.period = 30
        
        tnow = datetime.now()
        if self.stageSw:
            tstage1 = self.uptime + timedelta(seconds=self.stageInterval)
            tstage2 = self.uptime + timedelta(seconds=self.stageInterval*2)
            tstage3 = self.uptime + timedelta(seconds=self.stageInterval*3)
        else:
            tstage1 = self.uptime + timedelta(days=self.period)
            tstage2 = self.uptime + timedelta(days=self.period*2)
            tstage3 = self.uptime + timedelta(days=self.period*3)

        if tnow > tstage1 and self.stage < self.STAGE_1_B and self.rTime > STEP1_RECORD_TIME:
            self.stage = self.STAGE_1_B        
        elif tnow > tstage2 and self.stage < self.STAGE_2_B and self.rTime > STEP2_RECORD_TIME:
            self.stage = self.STAGE_2_B
        elif tnow > tstage3 and self.stage < self.STAGE_3_B and self.rTime > STEP3_RECORD_TIME and self.used >= STEP3_USED_TIME:
            self.stage = self.STAGE_3_B

        self.rmState.change_key_value(self.rmCom.SEC_STATE, self.rmCom.KEY_CLEARSTAGE, str(self.stage))

        logging.info("[RMCLEAR] update stage [%d].", self.stage)

    def clear_act(self):
        if self.flag:
            return
        self.flag = True
        logging.info("[RMCLEAR] clear act start.")
        if self.stage == self.STAGE_1_B:
            self.stage1_act()
        elif self.stage == self.STAGE_2_B:
            self.stage2_act()
        elif self.stage == self.STAGE_3_B:
            self.stage3_act()
        self.flag = False

    # 没有兼容在第一阶段就删掉fs.old
    def stage1_act(self):
        if self.stage != self.STAGE_1_B:
            logging.info("[RMCLEAR] In[%d], but this is stage [%d].", self.stage, self.STAGE_1_B)
            return
        logging.info("[RMCLEAR] Step 1 start.")
        update_grub()

        if self.osName == None:
            logging.error("[RMCLEAR] Step 1 oldfs not exists.")
            return

        if os.path.exists("/fs.old"):
            cmd  = "rm /fs.old"
            get_sys_call(cmd)

        whiteListPath = self.osName + STEP2_LIST_FILE
        if not os.path.exists(whiteListPath):
            cmd = "rm -rf " + self.osName
            get_sys_call(cmd)
            self.stage = self.STAGE_E
            self.set_clear_state(True)
            logging.error("[RMCLEAR] Step 1 [%s] not exists, clear all and is over.", whiteListPath)
            return

        if not os.path.exists(STEP1_LIST_FILE):
            logging.error("[RMCLEAR] [%s] not exists.", STEP1_LIST_FILE)
            return
        try:
            pathList = []
            with open(STEP1_LIST_FILE, 'r') as fd:
                pathList = fd.readlines()
            fd.close()
            
            for ele in pathList:
                path = self.osName + ele.strip("\n")
                if os.path.exists(path):
                    shutil.rmtree(path)
                    logging.debug("[RMCLEAR] Step 1 remove [%s].", path)
        except Exception as e:
            logging.error("[RMCLEAR] Step 1 err[%s].", e)

        downloadPath = self.rmState.get_key(self.rmCom.SEC_OS_INFO, self.rmCom.KEY_ISO_PATH)
        if downloadPath != None and len(downloadPath) > 0 and os.path.exists(downloadPath):
            logging.info("[RMCLEAR] remove iso[%s].", downloadPath)
            shutil.rmtree(downloadPath)

        self.stage = self.STAGE_1_E
        logging.info("[RMCLEAR] Step 1 end.")

    def stage2_act(self):
        if self.stage != self.STAGE_2_B:
            logging.info("[RMCLEAR] In[%d], but this is stage [%d].", self.stage, self.STAGE_2_B)
            return
        logging.info("[RMCLEAR] Step 2 start.")

        if self.osName == None:
            logging.error("[RMCLEAR] Step 2 oldfs not exists.")
            return
        # eg:
        # /.fs.old-v10   <-    /fs.old-v10
        if self.osName.startswith("/"):
            newFsDir = "/." + self.osName[1:]
        else:
            newFsDir = "/." + self.osName

        whiteListPath = self.osName + STEP2_LIST_FILE
        if not os.path.exists(whiteListPath):
            logging.error("[RMCLEAR] Step 2 [%s] not exists.", whiteListPath)
            return
        whiteList = []
        with open(whiteListPath, 'r') as fd:
            whiteList = fd.readlines()
        fd.close()
        for ele in whiteList:
            # eg: ele
            # /.
            # /lib/aarch64-linux-gnu/libreadline.so.6.3
            path = self.osName + ele.strip("\n")
            if not os.path.exists(path):
                logging.warning("[RMCLEAR] Step 2 whiteList [%s] not exists.", path)
                continue
            if os.path.isdir(path):
                logging.debug("[RMCLEAR] Step 2 whiteList [%s] is dir.", path)
                continue

            newPath = newFsDir + ele.strip("\n")
            # eg: newPath
            # /.fs.old-v10/lib/aarch64-linux-gnu/libreadline.so.6.3
            try:
                tmpDir = os.path.dirname(newPath)
                if not os.path.exists(tmpDir):
                    os.makedirs(tmpDir)
                cmd = "mv " + path + " " + newPath
                get_sys_call(cmd)
                # logging.debug("[RMCLEAR] Step 2 whiteList move [%s] to [%s].", path, newPath)
            except Exception as e:
                logging.error("[RMCLEAR] Step 2 whiteList move [%s] to [%s] err[%s].", path, newPath, e)

        tnow = datetime.now()
        timeD = tnow - self.uptime
        deltaMin = int(timeD.seconds / 60)
        logging.debug("[RMCLEAR] Step 2 atime delta minute [%d].", deltaMin)
        cmd = "find " + self.osName + " -amin -" + str(deltaMin)
        retInfo = get_sys_call(cmd)
        retList = retInfo.split("\n")
        for ele in retList:
            # eg:
            # /fs.old-v10
            # /fs.old-v10/lib/aarch64-linux-gnu/libreadline.so.6.3
            if len(ele) == 0:
                continue
            if not os.path.exists(ele):
                logging.warning("[RMCLEAR] Step 2 atime [%s] not exists.", ele)
                continue
            if os.path.isdir(ele):
                logging.debug("[RMCLEAR] Step 2 atime [%s] is dir.", ele)
                continue
            newPath = newFsDir + ele[len(self.osName):]
            try:
                tmpDir = os.path.dirname(newPath)
                if not os.path.exists(tmpDir):
                    os.makedirs(tmpDir)
                cmd = "mv " + ele + " " + newPath
                get_sys_call(cmd)
                # logging.debug("[RMCLEAR] Step 2 atime move [%s] to [%s].", ele, newPath)
            except Exception as e:
                logging.error("[RMCLEAR] Step 2 atime move [%s] to [%s] err[%s].", ele, newPath, e)

        if os.path.exists(TMP_FS):
            shutil.rmtree(TMP_FS)
            logging.warning("[RMCLEAR] Step 2 [%s] exists, so rm it.", TMP_FS)
        cmd = "mv " + self.osName + " " + TMP_FS
        ret = subprocess.run(cmd, shell=True, stdout=subprocess.PIPE)
        if ret.returncode != 0:
            logging.error("[RMCLEAR] Step 2 move [%s] to [%s] out[%s] err[%s].", self.osName, TMP_FS, ret.stdout.decode(), ret.stderr.decode())

        self.stage = self.STAGE_2_E
        logging.info("[RMCLEAR] Step 2 end.")


    def stage3_act(self):
        if self.stage != self.STAGE_3_B:
            logging.info("[RMCLEAR] In[%d], but this is stage [%d].", self.stage, self.STAGE_3_B)
            return
        logging.info("[RMCLEAR] Step 3 start.")

        if os.path.exists(TMP_FS):
            shutil.rmtree(TMP_FS)
            logging.info("[RMCLEAR] Step 3 remove [%s].", TMP_FS)
        else:
            logging.info("[RMCLEAR] Step 3 [%s] not exists.", TMP_FS)

        self.stage = self.STAGE_E
        self.set_clear_state(True)
        logging.info("[RMCLEAR] Step 3 end.")


    def manual_clear(self):
        if not self.is_up():
            logging.info("[RMCLEAR] Os is not upgraded.")
            return -1, _("Os is not upgraded.")
        if self.is_cleared():
            logging.info("[RMCLEAR] is cleared.")
            return 0, _("Oldfs is cleared.")
        if not self.get_oldfsName():
            logging.error("[RMCLEAR] can not get old fs name.")
            return -1, _("oldfs not exists")

        logging.info("[RMCLEAR] manual clear start.")
        self.set_switch(True)
        self.stage = self.STAGE_1_B
        self.rmState.change_key_value(self.rmCom.SEC_STATE, self.rmCom.KEY_CLEARSTAGE, str(self.stage))
        self.clear_act()
        if self.timer == None or not self.timer.is_alive():
            threadTmp = threading.Thread(target=self.auto_clear)
            threadTmp.start()

        return 0, _("Start clear.")

if __name__ == "__main__":
    loglevel = logging.INFO
    FORMAT = "%(asctime)s [%(levelname)s]: %(message)s"
    logFileName = "/var/log/RevisionsManager/RevisionsManager.log.1"
    logging.basicConfig(format=FORMAT, level=loglevel, datefmt='%m-%d,%H:%M:%S', filename=logFileName, filemode='a')
    obj = RMClear()
    sleep(7)
    obj.manual_clear()

